package com.me.asteroids.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.math.Vector2;
import com.me.asteroids.Constants;
import com.me.asteroids.EntityFactory;
import com.me.asteroids.Graphics;
import com.me.asteroids.components.AccelerationComponent;
import com.me.asteroids.components.AsteroidComponent;
import com.me.asteroids.components.BulletComponent;
import com.me.asteroids.components.ColliderComponent;
import com.me.asteroids.components.DebrisComponent;
import com.me.asteroids.components.DecayComponent;
import com.me.asteroids.components.GameDataComponent;
import com.me.asteroids.components.ModelComponent;
import com.me.asteroids.components.PlayerComponent;
import com.me.asteroids.components.PositionComponent;
import com.me.asteroids.components.VelocityComponent;
import com.me.asteroids.components.model.PolygonModel;
import com.me.asteroids.events.AsteroidHitEvent;
import com.me.asteroids.events.BulletAsteroidCollisionEvent;
import com.me.asteroids.events.PlayerASteroidCollisionEvent;
import com.me.asteroids.events.PlayerDeathEvent;
import com.me.asteroids.events.ScreenWrapEvent;
import com.me.asteroids.systems.AsteroidSpawningSystem;
import com.me.asteroids.systems.CollisionSystem;
import com.me.asteroids.systems.DecaySystem;
import com.me.asteroids.systems.GameDataRenderSystem;
import com.me.asteroids.systems.GameDataSystem;
import com.me.asteroids.systems.ModelRenderSystem;
import com.me.asteroids.systems.MovementSystem;
import com.me.asteroids.systems.PlayerInputSystem;
import com.me.asteroids.systems.ScreenWrapSystem;
import com.me.common.Screen;
import com.me.common.ecs.ComponentMapper;
import com.me.common.ecs.Engine;
import com.me.common.ecs.Entity;
import com.me.common.ecs.event.EventHandler;
import com.me.common.ecs.event.Listener;

public class GameScreen extends Screen implements Listener {

    Engine engine;

    Graphics graphics;

    SpriteBatch batch;
    BitmapFont font;

    public GameScreen(Graphics graphics) {
        this.graphics = graphics;
    }

    @Override
    public void setup() {
        batch = graphics.getUISpriteBatch();
        font = new BitmapFont(Gdx.files.internal("Retro_Font.fnt"));

        engine = new Engine();

        engine.registerComponentClass(GameDataComponent.class);
        engine.registerComponentClass(PlayerComponent.class);
        engine.registerComponentClass(BulletComponent.class);
        engine.registerComponentClass(AsteroidComponent.class);
        engine.registerComponentClass(DebrisComponent.class);
        engine.registerComponentClass(DecayComponent.class);
        engine.registerComponentClass(ColliderComponent.class);
        engine.registerComponentClass(PositionComponent.class);
        engine.registerComponentClass(VelocityComponent.class);
        engine.registerComponentClass(AccelerationComponent.class);
        engine.registerComponentClass(ModelComponent.class);

        GameDataSystem system = new GameDataSystem(engine);

        engine.registerSystem(system);
        engine.registerSystem(new PlayerInputSystem(engine));
        engine.registerSystem(new AsteroidSpawningSystem(engine));
        engine.registerSystem(new DecaySystem(engine));
        engine.registerSystem(new MovementSystem(engine));
        engine.registerSystem(new CollisionSystem(engine));
        engine.registerSystem(new ScreenWrapSystem(engine));
        engine.registerSystem(new ModelRenderSystem(engine, graphics));
        engine.registerSystem(new GameDataRenderSystem(engine, graphics, font));

        engine.registerListener(system);
        engine.registerListener(this.new EventListener(engine));

        engine.ready();

        Entity gameData = engine.createEntity();
        gameData.addComponent(new GameDataComponent());
        gameData.activate();

        Entity player = EntityFactory.createPlayer(engine);
        player.activate();
    }

    @Override
    public void update(float dt) {
        graphics.reset();
        engine.update(dt);

        if (Constants.DEBUG) {
            batch.begin();
            font.setColor(Color.RED);
            font.draw(batch, String.format("FPS: %d, Entities: %d", Gdx.graphics.getFramesPerSecond(), engine.getEntityCount()), 15, 15 + font.getCapHeight());
            batch.end();
        }
    }

    @Override
    public void dispose() {
        font.dispose();
    }

    private class EventListener implements Listener {

        ComponentMapper<PositionComponent> positionMapper;
        ComponentMapper<ModelComponent> modelMapper;
        ComponentMapper<BulletComponent> bulletMapper;
        ComponentMapper<DebrisComponent> debrisMapper;
        ComponentMapper<AsteroidComponent> asteroidMapper;
        ComponentMapper<PlayerComponent> playerMapper;

        EventListener(Engine engine) {
            this.positionMapper = engine.getComponentMapper(PositionComponent.class);
            this.modelMapper = engine.getComponentMapper(ModelComponent.class);
            this.bulletMapper = engine.getComponentMapper(BulletComponent.class);
            this.debrisMapper = engine.getComponentMapper(DebrisComponent.class);
            this.asteroidMapper = engine.getComponentMapper(AsteroidComponent.class);
            this.playerMapper = engine.getComponentMapper(PlayerComponent.class);
        }

        @EventHandler
        public void onScreenWrap(ScreenWrapEvent event) {
            if (bulletMapper.has(event.entity) || debrisMapper.has(event.entity)) {
                // Remove bullets when they leave the screen
                event.setCancelled(true);
                event.entity.remove();
            }
        }

        @EventHandler
        public void onBulletAsteroidCollision(BulletAsteroidCollisionEvent event) {
            Vector2 bulletPosition = positionMapper.get(event.getBullet()).position;

            PolygonModel model = (PolygonModel) modelMapper.get(event.getAsteroid()).model;
            if (model.contains(bulletPosition)) {
                // AABBs intersect but let's only consider it a hit if the bullet's position
                // is actually inside the asteroid
                AsteroidComponent asteroid = asteroidMapper.get(event.getAsteroid());
                engine.callEvent(new AsteroidHitEvent(asteroid));
                event.getBullet().remove();
                if (asteroid.generation < 2) {
                    for (Entity shard : EntityFactory.splitAsteroidIntoChunks(engine, event.getAsteroid(), 2, 2/3f)) {
                        shard.activate();
                    }
                } else {
                    for (Entity debris : EntityFactory.createDebris(engine, event.getAsteroid())) {
                        debris.activate();
                    }
                }
                event.getAsteroid().remove();
            }
        }

        @EventHandler
        public void onPlayerAsteroidCollision(PlayerASteroidCollisionEvent event) {
            PolygonModel asteroid = (PolygonModel) modelMapper.get(event.getAsteroid()).model;
            PolygonModel player = (PolygonModel) modelMapper.get(event.getPlayer()).model;

            if (asteroid.contains(player.getVertices()) || player.contains(asteroid.getVertices())) {
                engine.callEvent(new PlayerDeathEvent(event.getPlayer()));
                playerMapper.get(event.getPlayer()).afterBurner.deactivate();
                event.getPlayer().deactivate();
                for (Entity debris : EntityFactory.createDebris(engine, event.getPlayer())) {
                    debris.activate();
                }
            }
        }

    }


}
